unit FormMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, ComCtrls, Menus, ToolWin, Gta3Archive, StdCtrls,
  ActnList, ShellApi, ShlObj, ImgList, FormAbout, FormProgress, History;

type

  TfrmMain = class(TForm)
    mnuMain: TMainMenu;
    stbMain: TStatusBar;
    lsvFileList: TListView;
    mniFile: TMenuItem;
    mniNew: TMenuItem;
    mniOpen: TMenuItem;
    mniClose: TMenuItem;
    mniN1: TMenuItem;
    mniExit: TMenuItem;
    dlgOpen: TOpenDialog;
    ActionList1: TActionList;
    actClose: TAction;
    mniAction: TMenuItem;
    actExtract: TAction;
    mniExtract: TMenuItem;
    dlgSave: TSaveDialog;
    mniAdd: TMenuItem;
    mniDelete: TMenuItem;
    actAdd: TAction;
    actDelete: TAction;
    dlgAdd: TOpenDialog;
    imlSmall: TImageList;
    imlLarge: TImageList;
    mniHelp: TMenuItem;
    mniAbout: TMenuItem;
    actSelectAll: TAction;
    mniN2: TMenuItem;
    SelectAll1: TMenuItem;
    ToolBar1: TToolBar;
    actRename: TAction;
    mniRename: TMenuItem;
    mniReopen: TMenuItem;
    History: THistory;
    PopupMenu1: TPopupMenu;
    Add1: TMenuItem;
    Extract1: TMenuItem;
    Rename1: TMenuItem;
    Delete1: TMenuItem;
    N1: TMenuItem;
    SelectAll2: TMenuItem;
    procedure lsvFileListData(Sender: TObject; Item: TListItem);
    procedure lsvFileListDataFind(Sender: TObject; Find: TItemFind;
      const FindString: String; const FindPosition: TPoint;
      FindData: Pointer; StartIndex: Integer; Direction: TSearchDirection;
      Wrap: Boolean; var Index: Integer);
    procedure mniOpenClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure lsvFileListColumnClick(Sender: TObject; Column: TListColumn);
    procedure actCloseUpdate(Sender: TObject);
    procedure actCloseExecute(Sender: TObject);
    procedure mniNewClick(Sender: TObject);
    procedure actDeleteUpdate(Sender: TObject);
    procedure mniExitClick(Sender: TObject);
    procedure actExtractUpdate(Sender: TObject);
    procedure actAddExecute(Sender: TObject);
    procedure actExtractExecute(Sender: TObject);
    procedure actDeleteExecute(Sender: TObject);
    procedure mniAboutClick(Sender: TObject);
    procedure actSelectAllExecute(Sender: TObject);
    procedure lsvFileListChange(Sender: TObject; Item: TListItem;
      Change: TItemChange);
    procedure actRenameUpdate(Sender: TObject);
    procedure actRenameExecute(Sender: TObject);
    procedure HistoryChange(Sender: TObject);
    procedure GenericReopenClick(Sender: TObject);
  private
    { Private declarations }
    FCurrentItem: Integer;
    ViewForm: TForm;
    procedure OpenArchive(FileName: string);
    procedure CloseArchive;
    procedure ExtractEntry(Index: Integer; FileName: string);
    procedure ExtractSelected(OutDir: string);
    procedure ExtractAll(OutDir: string);
    procedure AddFiles(Files: TStrings);
    function BrowseForFolder(InitialDir: String): String;
    procedure ViewEntry(Index: Integer);
  public
    { Public declarations }
    function GetTypeName(Name: string): string;
    function GetTypeIndex(Name: string): Integer;
  end;

const
  APP_NAME = 'ImgEd';
  APP_VER  = '0.2';

var
  frmMain: TfrmMain;
  Archive: TGTA3Archive;
  Sort : record
    Func: Pointer;
    Direction:integer;
  end;

implementation

uses SortFunctions;

{$R *.dfm}

procedure TfrmMain.OpenArchive(FileName: string);
begin
  CloseArchive;
  Archive := TGTA3Archive.Create(FileName);
  Caption := APP_NAME + ' - [' + Archive.ArchiveName + ']';
  Archive.Sort(Sort.Func);
  lsvFileList.Items.Count := Archive.EntryCount;
  FCurrentItem := -1;
  lsvFileList.Repaint;
  History.AddItem(FileName);
end;

procedure TfrmMain.CloseArchive;
begin
  lsvFileList.Items.Count := 0;
  lsvFileList.Repaint;
  if Assigned(Archive) then
    FreeAndNil(Archive);
  Caption := APP_NAME;
end;

procedure TfrmMain.lsvFileListData(Sender: TObject; Item: TListItem);
begin
  if Archive <> nil then
  begin
    Item.Caption := Trim(Archive.Entry[Item.Index].Name);
    Item.ImageIndex := GetTypeIndex(Item.Caption);
    if lsvFileList.ViewStyle = vsReport then
    begin
      Item.SubItems.Add(GetTypeName(Item.Caption));
      Item.SubItems.Add(IntToStr(Archive.Entry[Item.Index].BlockCount * 2) + ' kB');
    end;
  end;
end;

procedure TfrmMain.lsvFileListDataFind(Sender: TObject; Find: TItemFind;
  const FindString: String; const FindPosition: TPoint; FindData: Pointer;
  StartIndex: Integer; Direction: TSearchDirection; Wrap: Boolean;
  var Index: Integer);
var
  I: Integer;
  Found: Boolean;
begin
  I := StartIndex;
  if (Find = ifExactString) or (Find = ifPartialString) then
  begin
    repeat
      if (I = Archive.EntryCount) then
        if Wrap then I := 0 else Exit;
      Found := Pos(UpperCase(FindString), UpperCase(Archive.Entry[I].Name)) = 1;
      Inc(I);
    until Found or (I = StartIndex);
    if Found then Index := I-1;
  end;
end;

procedure TfrmMain.mniOpenClick(Sender: TObject);
begin
  if dlgOpen.Execute then
    OpenArchive(dlgOpen.FileName);
end;

procedure TfrmMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  CloseArchive;
  History.SaveToIni(IncludeTrailingBackslash(ExtractFilePath(Application.ExeName)) + 'ImgEd.ini');
end;

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  Sort.Func := @SortByName;
  Sort.Direction := SORT_ASCENDING;
  Application.Title := APP_NAME;
  History.LoadFromIni(IncludeTrailingBackslash(ExtractFilePath(Application.ExeName)) + 'ImgEd.ini');
  if ParamCount > 0 then
    OpenArchive(ParamStr(1));
end;

procedure TfrmMain.lsvFileListColumnClick(Sender: TObject;
  Column: TListColumn);
var
  P: TListSortCompare;
begin
  if Archive = nil then Exit;
  
  P := Sort.Func;
  if Column.Index = 0 then
    Sort.Func := @SortByName
  else if Column.Index = 1 then
    Sort.Func := @SortByType
  else if Column.Index = 2 then
    Sort.Func := @SortBySize;

  if @P = Sort.Func then
    Sort.Direction := Sort.Direction xor 1
  else
    Sort.Direction := SORT_ASCENDING;

  Archive.Sort(Sort.Func);
  lsvFileList.Repaint;
end;

procedure TfrmMain.ExtractEntry(Index: Integer; FileName: string);
var f: TFileStream;
begin
  f := TFileStream.Create(FileName, fmCreate);
  Archive.Extract(Index, f);
  f.Free;
end;

procedure TfrmMain.ExtractSelected(OutDir: string);
var i: Integer;
begin
  with TfrmProgress.Create(Application) do
  try
    SetTitle('Extracting File(s)...');
    SetProgressPos(0);
    SetProgressMax(lsvFileList.SelCount);
    Show;
    OutDir := IncludeTrailingBackslash(OutDir);
    for i := 0 to lsvFileList.Items.Count -1 do
    begin
      if lsvFileList.Items[i].Selected then
      begin
        SetSubTitle(Trim(Archive.Entry[i].Name));
        ProgressStepIt;
        ExtractEntry(i, OutDir + Trim(Archive.Entry[i].Name));
      end;
    end;
  finally
    Free;
  end;
end;

procedure TfrmMain.ExtractAll(OutDir: string);
var i: Integer;
begin
  OutDir := IncludeTrailingBackslash(OutDir);
  for i := 0 to lsvFileList.Items.Count -1 do
  begin
    ExtractEntry(i, OutDir + Trim(Archive.Entry[i].Name));
  end;
end;

function TfrmMain.GetTypeName(Name: string): string;
var
  Tmp: string;
begin
  Tmp := Uppercase(ExtractFileExt(Name)) + ' File';
  Result := Copy(Tmp, 2, Length(Tmp)-1);
end;

function TfrmMain.GetTypeIndex(Name: string): Integer;
var Ext: string;
begin
  Result := 0;
  Ext := LowerCase(ExtractFileExt(Name));
  if Ext = '.txd' then
    Result := 1
  else if Ext = '.dff' then
    Result := 2
  else
end;

procedure TfrmMain.actCloseUpdate(Sender: TObject);
begin
  (Sender as TAction).Enabled := Assigned(Archive);
end;

procedure TfrmMain.actCloseExecute(Sender: TObject);
begin
  CloseArchive;
end;

procedure TfrmMain.mniNewClick(Sender: TObject);
begin
  if dlgSave.Execute then
  begin
    CloseArchive;
    Archive := TGTA3Archive.Create(dlgSave.FileName, True);
    Caption := APP_NAME + ' - [' + Archive.ArchiveName + ']'; 
  end;
end;

procedure TfrmMain.actDeleteUpdate(Sender: TObject);
begin
  (Sender as TAction).Enabled := Assigned(Archive) and (lsvFileList.SelCount > 0);
end;

procedure TfrmMain.mniExitClick(Sender: TObject);
begin
  Close;
end;

procedure TfrmMain.actExtractUpdate(Sender: TObject);
begin
  (Sender as TAction).Enabled := Assigned(Archive) and (lsvFileList.SelCount > 0);
end;

procedure TfrmMain.actAddExecute(Sender: TObject);
begin
  if dlgAdd.Execute then
  begin
    AddFiles(dlgAdd.Files);
    FCurrentItem := -1;
  end;
end;

procedure TfrmMain.AddFiles(Files: TStrings);
var
  i: integer;
  f: TFileStream;
begin
  Screen.Cursor := crHourGlass;
  try
    with TfrmProgress.Create(Application) do
    try
      SetTitle('Adding File(s)...');
      SetProgressPos(0);
      SetProgressMax(Files.Count);
      Show;
      for i := 0 to Files.Count -1 do
      begin
        f := TFileStream.Create(Files.Strings[i], fmShareDenyNone);
        Archive.Add(ExtractFileName(Files.Strings[i]), f);
        ProgressStepIt;
      end;
      Archive.Sort(Sort.Func);
      lsvFileList.Items.Count := Archive.EntryCount;
      lsvFileList.Repaint;
    finally
      Free;
    end;
  finally
    Screen.Cursor := crDefault;
  end;
end;

procedure TfrmMain.actExtractExecute(Sender: TObject);
var
  Dir: string;
begin
  Dir := BrowseForFolder('');
  if Dir <> '' then
  begin
    ExtractSelected(IncludeTrailingBackslash(Dir));
  end;
end;

procedure TfrmMain.actDeleteExecute(Sender: TObject);
var i: integer;
begin
  if MessageDlg('Are you sure you want to delete the selected files?',
    mtConfirmation, [mbYes, mbNo], 0) = mrYes then
  begin
    try
      for i := 0 to lsvFileList.Items.Count -1 do
      begin
        if lsvFileList.Items[i].Selected then
        begin
          Archive.MarkForRemoval(i);
        end;
      end;
    finally
      Archive.DoRemove;
      FCurrentItem := -1;
      lsvFileList.Items.Count := Archive.EntryCount;
      //We need to resort because 'DoRemove' sorts it by 'index'
      Archive.Sort(Sort.Func);
      lsvFileList.ClearSelection;
      lsvFileList.Repaint;
    end;
  end;
end;

// Used by BrowseForFolder to set the initial dir
function SetSelectionBrowseCallback(hwnd: HWND; uMsg: LongWord; lParam: LongInt;
  lpData: LongInt): Integer; stdcall;
begin
  if (uMsg = BFFM_INITIALIZED) then
  begin
    SendMessage(hwnd, BFFM_SETSELECTION, 1, lpData);
  end;
end;

function TfrmMain.BrowseForFolder(InitialDir: String): String;
var
  TitleName : string;
  lpItemID : PItemIDList;
  BrowseInfo : TBrowseInfo;
  DisplayName : array[0..MAX_PATH] of char;
  TempPath : array[0..MAX_PATH] of char;
begin
  result := '';
  FillChar(BrowseInfo, sizeof(TBrowseInfo), #0);
  BrowseInfo.hwndOwner := Handle;
  BrowseInfo.pszDisplayName := @DisplayName;
  TitleName := 'Please specify a directory';
  BrowseInfo.lpszTitle := PChar(TitleName);
  BrowseInfo.ulFlags := BIF_RETURNONLYFSDIRS;

  if InitialDir <> '' then
  begin
    BrowseInfo.lpfn := @SetSelectionBrowseCallback;
    BrowseInfo.lParam := integer(InitialDir);
  end;

  lpItemID := SHBrowseForFolder(BrowseInfo);
  if lpItemId <> nil then begin
    SHGetPathFromIDList(lpItemID, TempPath);
    Result := TempPath;
    GlobalFreePtr(lpItemID);
  end;
end;

procedure TfrmMain.mniAboutClick(Sender: TObject);
begin
  with TfrmAbout.Create(Application) do
  try
    ShowModal;
  finally
    Free;
  end;
end;

procedure TfrmMain.actSelectAllExecute(Sender: TObject);
begin
  lsvFileList.SelectAll;
end;

procedure TfrmMain.lsvFileListChange(Sender: TObject; Item: TListItem;
  Change: TItemChange);
var
  F: TMemoryStream;
begin
{
  if Archive = nil then Exit;
  if Item = nil then exit;
  if Change <> ctState then Exit;
  if not Item.Selected then Exit;
  // Prevent current model to be reloaded an extra time before loeding the now
  // one.
  if FCurrentItem = Item.Index then
  begin
    FCurrentItem := -1;
    Exit;
  end;

  ViewEntry(Item.Index);
}
end;

procedure TfrmMain.ViewEntry(Index: Integer);
begin
  //TODO
end;

procedure TfrmMain.actRenameUpdate(Sender: TObject);
begin
  (Sender as TAction).Enabled := Assigned(Archive) and (lsvFileList.SelCount = 1);
end;

procedure TfrmMain.actRenameExecute(Sender: TObject);
var
  Tmp: string;
begin
  Tmp := InputBox('Rename file...', '?', Archive.Entry[lsvFileList.Selected.Index].Name);
  if (Tmp <> '') and (Tmp <> Archive.Entry[lsvFileList.Selected.Index].Name) then
  begin
    Archive.Rename(lsvFileList.Selected.Index, Tmp);
    lsvFileList.Repaint;
  end;
end;

procedure TfrmMain.HistoryChange(Sender: TObject);
var i: integer; 
begin 
  with mniReopen do
  begin 
    Clear; 
    for i:= 0 to History.Count-1 do
    begin 
      Add(TMenuItem.Create(self)); 
      with Items[Count-1] do 
      begin 
        Caption := '&' + IntToStr(i) +' '+ History.Items[i];
        Tag := i; 
        OnClick := GenericReopenClick; 
      end; 
    end; 
    Enabled := Count > 0; 
    // Add linebreak 
    Add(TMenuItem.Create(self)); 
    Items[Count-1].Caption := '-'; 
    // Add 'Clear History' 
    Add( TMenuItem.Create(self) ); 
    with Items[Count-1] do 
    begin 
      Caption := 'Clear History'; 
      Tag := -1; 
      OnClick := GenericReopenClick; 
    end; 
  end; 
end;

procedure TfrmMain.GenericReopenClick(Sender: TObject);
begin 
  if TMenuitem(Sender).Tag >= 0 then
    OpenArchive(History.Items[TMenuitem(Sender).Tag])
  else
    History.Clear;
end;

end.
